/** A mapping from A to B. */
interface MapI<A,B> {
  MapI<A,B> newMap();
  B get(A key);
  B add(A key, B value);
  boolean contains(A key);
  /** Returns the number of keys in this (and hence the number of pairs (a, b) in this). */
  int size();
//  /** Returns the set of keys in this map. */
//  SetI<A> getKeys();
}

/** A map over an ordered domain A. */
interface OrderedMap<A extends Comparable<A>, B> extends Map<A, B> {
//  /* Narrows narrows output type of inherited method. */
//  OrderedSetI<A> getKeys();
}

/** A mutable binary search tree of elements of type T.  Assume no duplicate items (of type T).  
  * This interface can easily implemented as a singly linked list. */
interface TreeMapI<A, B> extends MapI { 
  /** Returns the length of this list. */
  int size();
  /** Returns the value for given key; null if this does not contain key. */
  B get(A key);
  /** Adds the (key, value) pair to the table, replacing the former value for key if such exists. 
    * @ returns true if key was already present in the map; false otherwise. */
  boolean add(A key, B value);
  
  /** Inserts the object o in this Tree, assuming it does not already exist. */
  BiListI<T> delete(T o);      
  /** The inherited newIterator with a narrowed output type */
  BiIteratorI<T> newIterator();
}

/* A record including an embedded key which supports the Comparable<K> interface.  The ordering on Ordered<T,K>
 * is induced by the Comparable ordering on K. */
interface Ordered<K extends Comparable<K>, T extends Ordered<K,T>> extends Comparable<T> {
  int compareTo(T other);
  boolean match(K key);
}

/** An iterator that traverses a collection of objects of type T (such as an EnumI<T>) possibly modifying it. */
interface IteratorI<T> extends ReadIteratorI<T> {
  
  /** Destructively inserts the object o in the search tree corresponding to this iterator. */
  void insert(T o);
  
  /** Removes the current item.
    * @throws an IteratorException if the cursor is atEnd(). */
  T remove(); 
}

/** An iterator that traverses a collection of objects of type T in either direction possibly modifying it. */
interface BiIteratorI<T> extends IteratorI<T> {  // T implicitly bound by Ordered<T>
  /** Places the cursor on the last element of the collection.  The behavior is unspecified if the collection is empty. */
  T last(); 
  /** Places the cursor on the prevoius element of the collection.  The behavior is unspecified if the collection is empty. */
  T prev(); 
}

/** Exception class for signalling errors in processing lists (implementing ListI) */
class TreeException extends RuntimeException {
  TreeException(String s) { super(s); }
}

/** Exception class for signalling errors in using an iterator (implementing IteratorI) */
class IteratorException extends RuntimeException {
  IteratorException(String s) { super(s); }
}

/** A circularly linked class implementing the BiListI interface. */
class BinarySearchTree<T,K> implements BinarySearchTreeI<T,K> {
  
  /** The root node for the tree; may be null. */
  private Node<T> root = null;
  /** The size of this tree. */
  private int size = 0;
  
  /* Relying on the default constructor for BinarySearchTree. */
  
  public String toString(){
    StringBuffer s = new StringBuffer("(");
    for (IteratorI i = newIterator(); ! i.atEnd(); i.next()) {
      if (i.atStart()) s.append(i.currentItem().toString());
      else s.append(" " + i.currentItem().toString());
    }  
    return s.append(")").toString();
  }
 
  /* Implementations of BiListI methods.  See BiListI for contracts. */
  public boolean isEmpty() { return root == null } 
  public BinarySearchTreeI<T> newBinarySearchTree() { return new BinarySearchTree<T>(); }
  public int size() { return size; }
  
  public BiListI<T> insert(T o) { /** stub */ return this; }
  
  /* Virtual constructors for iterators; identical except for return type */
  public BiIteratorI<T> newIterator() { return new BiIterator<T>(); }
  
  boolean delete(T elt) {
    for (IteratorI<T> i = newIterator(); !i.atEnd(); i.next())
      if (i.currentItem().equals(elt)) {
      i.remove();
      return true;
    };
    return false;
  }
  
  /* Static nested Node class.  Embedded as nested class to hide access. */
  private static class Node<T> {
    T item;
    Node<T> pred,succ;
    
    Node(T i, Node<T> p, Node<T> s) {
      item = i;
      pred = p;
      succ = s;
    }
    
    Node() {  // allocate header
      item = null;
      pred = this;
      succ = this;
    }
  } 
  
  /** A true inner class that includes a link (BiList.this) to the enclosing BiList. */
  private class BiIterator implements BiIteratorI<T> {
    
    Node<T> current;
    
    BiIterator() {
      current = BiList.this.head.succ;  // current is first item (if one exists)
    }
    
    public T first() { 
      current = BiList.this.head.succ;  // current is first item (if one exists)
      return current.item;
    }
    
    public T last() { 
      current = BiList.this.head.pred;  // current is last item (if one exists)
      return current.item;
    }
    
    public T next() {
      current = current.succ;        // wraps around end
      return current.item;
    }
    
    public T prev() {
      current = current.pred;        // wraps around end
      return current.item;
    }
    
    public T currentItem() { 
      if (current == BiList.this.head) throw
        new IteratorException("No current element in " + BiList.this);
      return current.item; 
    }
    
    public boolean atEnd() { return current == BiList.this.head; }
   
    /** Returns true if CircList is non-empty and current is the first node in the list. */
    public boolean atStart() { return ! atEnd() && current == BiList.this.head.succ; }
    
    
    public void insert(T o) {
      
      // pre: true
      // post: Node containing o is inserted before current item, 
      //       current is unchanged
      
      Node<T> oldPred = current.pred;
      Node<T> newNode = new Node<T>(o, oldPred, current);  // allocate new node
      current.pred = newNode;                              // insert it
      oldPred.succ = newNode;
      BiList.this.length++;
      
    } 
    
    public T remove() {
      
      // pre:  current is valid
      // post: current becomes current.succ
      
      if (current == BiList.this.head) throw 
        new IteratorException("BiIterator.remove() applied at end of BiList " + BiList.this);
      Node<T> cPred = current.pred;
      Node<T> cSucc = current.succ;
      cPred.succ = cSucc;
      cSucc.pred = cPred;
      T remOb = current.item;
      current = cSucc;  
      BiList.this.length--;
      return remOb;
    } 
  }
  
  class BiListException extends RuntimeException {
    BiListException(String s) { super(s); }
  }
}

class TestBiList {
  public static void main(String args[]) {
    BiList<Integer> c = new BiList<Integer>();
    System.out.println(c.isEmpty());
    System.out.println("c is: " + c);
    c.insertFront(new Integer(567));
    System.out.println("c is: " + c);
    System.out.println(c.isEmpty());
    c.insertRear(new Integer(72));
    System.out.println("c is: " + c);
    c.insertRear(new Integer(4330));
    System.out.println("c is: " + c);
    BiList<BiList<Integer>> d = new BiList<BiList<Integer>>();
    d.insertFront(c);
    System.out.println("d is: " + d);
  }
}