/** A mapping from K to V. Null keys are not allowed. */
interface MapI<K,V> {
  
  /** create a new empty map of the same type. */
  MapI<K,V> newMap();
  
  /** @return true if the map is empty (contains no key-value pairs). */
  boolean isEmpty();
  
  /** Set this map to the empty map. */
  void clear();
  
  /** @return the value for key if there is an entry (key,value) in the map; null otherwise. */
  V get(K key);
  
  /** Puts the entry (key, value) in the map, replacing the old entry for key if one existed.
    * @return the old value for key or null if no old value existed in the map. */
  V put(K key, V value);
  
  /* Deletes the entry for key.
   * @return the value of the removed entry if such an entry existed, null otherwise. */
  V remove(K key);
  
  /* return true iff an entry for key exists in the map. */
  public boolean containsKey(K key);
  
  /** Returns the number of entries (keys) in this map. */
  int size();
}

/** A map over a totally-ordered domain K. */
interface SortedMapI<K extends Comparable<K>, V> extends MapI<K, V> {
  K firstKey();
  K lastKey();
}

/** A mutable binary search tree of entries of type (K,V). */ 
class TreeMap<K extends Comparable<K>, V> implements SortedMapI<K,V> {
   
  /** Number of entries in the map. */
  private int size;
  
  /** Root of the binary search tree.  Weak form of state pattern: either empty or node. 
    * An empty tree is makeEmpty == new RefNode(null).  A node is RefNode(n) where n is a Node. */
  RefNode<K,V> root;
  
  /* Constructors */
  TreeMap() { 
    root = makeEmpty();
    size = 0;
  }
  
  TreeMap(K key, V val) {
    root = makeLeaf(key, val);
    size = 1;
  }
  
  private TreeMap(RefNode<K,V> rn, int s) {
    root = rn;
    size = s;
  }
  
  /** @return a balanced TreeMap containing the elements in the items array. */
  static <K extends Comparable<K>> TreeMap<K,Void> makeBalancedTreeMap(K[] items) {     
    return new TreeMap<K,Void>(makeBalancedTree(items, 0, items.length - 1), items.length);   
  } 
  
  /** Factory method for constructing empty nodes. */
  private static <K extends Comparable<K>, V> RefNode<K,V> makeEmpty() { return new RefNode<K,V>(null); } 

  /** Factory method for constructing leaf nodes. */
  private static <K extends Comparable<K>, V> RefNode<K,V> makeLeaf(K key, V val) { return new RefNode<K,V>(key, val); }
  
  /** Assumes 0 <= lo <= hi < items.length;   
    * @return a balanced TreeMap containing the elements in the [lo:hi] slice of the items array, i.e., 
    * items[lo], ..., items[hi].  The value fields for the elements in the TreeMap are null.   */
  private static <K extends Comparable<K>> RefNode<K,Void> makeBalancedTree(K[] items, int lo, int hi) {
    int mid = (lo + hi)/2;
    RefNode<K,Void> newRoot = makeLeaf(items[mid], null);
    RefNode<K,Void> left = (mid - lo == 0) ? TreeMap.<K,Void>makeEmpty() : makeBalancedTree(items, lo, mid - 1);
    RefNode<K,Void> right = (hi - mid == 0) ? TreeMap.<K,Void>makeEmpty() : makeBalancedTree(items, mid + 1, hi);
    newRoot.setLeft(left);
    newRoot.setRight(right);
    return newRoot;
  }
  
  /** create a new empty map of the same type. */
  public MapI<K,V> newMap() { return new TreeMap<K,V>(); }
  
  /** @return true if the map is empty (contains no key-value pairs). */
  public boolean isEmpty() { return root.isEmpty(); }
  
  /** Set this map to the empty map. */
  public void clear() { 
    root = makeEmpty();
    size = 0;
  }
  
  /** @return the value for key if there is an entry (key,value) in the map; null otherwise. */
  public V get(K key) { return find(key).value(); }  // returns null if find(key) is empty
  
  /** Puts the entry (key, value) in the map, replacing the old entry for key if one existed.
    * @return the old value for key or null if no old value existed in the map.
    */
  public V put(K k, V v) {
    // location is the matching node which may be empty
    RefNode<K,V> location = find(k);  // may return empty
    V oldValue = location.value();    // may return null
    if (location.isEmpty()) {  // key k does not yet exist in tree
      location.set(k,v);
      size++;
    }
    else location.setValue(v);  // key k already exists in tree
    return oldValue;
  }
  
  /* Deletes the entry for key.
   * @return the value of the removed entry if such an entry existed, null otherwise. */
  public V remove(K key) {
    RefNode<K,V> location = find(key);  // may return root
    if (location.isEmpty()) return null; // no match found
    size--;
    return location.remove();  // returns old value in location
  } 
  
  /** Finds the RefNode matching key, which is an empty location if ! containsKey(key) 
    * @returns the matching RefNode, which may be empty */
  private RefNode<K,V> find(K key) {
    RefNode<K,V> current = root;
    while (! current.isEmpty()) {
      int compareFlag = key.compareTo(current.key());
      if (compareFlag == 0) break;
      else if (compareFlag < 0) current = current.left();
      else current = current.right();
    }
    return current;
  }
  
  /** Requires: ! root.isEmpty()
    * @returns the first non-empty RefNode in this.
    * @throws MapException if this is empty. */
  private RefNode<K,V> findFirst() {
    if (isEmpty()) throw new MapException("Attempted to find first node of empty TreeMap");
    RefNode<K,V> current = root;
     RefNode<K,V> next = root.left();
     while (true) {
       if (next.isEmpty()) return current;
       current = next; next = next.left();
     }     
  }
  
  /** Requires: ! root.isEmpty()
    * @returns the last non-empty RefNode in this.
    * @throws MapException if this is empty. */
  private RefNode<K,V> findLast() {
    if (isEmpty()) throw new MapException("Attempted to find first node of empty TreeMap");
    RefNode<K,V> current = root;
    RefNode<K,V> next = root.right();     
    while (true)  {
      if (next.isEmpty()) return current;
      current = next;  next = next.right();
    }
  }
  
  /** Requires: ! root.isEmpty()
    * @returns the first key in this.
    * @throws MapException if this is empty. */
  public K firstKey() { return findFirst().key(); }
  
  /** Requires: ! root.isEmpty()
    * @returns the last key.
    * @throws MapException if this is empty. */
  public K lastKey() { return findLast().key(); }
    
  /* return true iff an entry for key exists in the map. */
  public boolean containsKey(K key) { return get(key) != null; }
  
  /** Returns the size of this map. */
  public int size() { return size; }
  
  public String toString() {
    if (isEmpty()) { return "{}"; }
    
    StringBuilder sb = new StringBuilder();
    toStringHelp(sb, root);
    
    /* sb has a leading space */
    sb.setCharAt(0, '{');
    sb.append('}');
    return sb.toString();
  }
  
  private void toStringHelp(StringBuilder sb, RefNode<K,V> n) {
    if (! n.isEmpty()) {
      toStringHelp(sb, n.left());
      sb.append(" [" + n.key() + "," + n.value() + "]");
      toStringHelp(sb, n.right());
    }
  }
  
  /** A reference class (box) representing an updateable cell that holds a Node. */
  private static class RefNode<K extends Comparable<K>,V> {
    Node<K,V> node;
    
    /** Constructors */
    RefNode() { this(null); }
    RefNode(K key, V val) { this(new Node<K,V> (key, val)); }
    
    private RefNode(Node<K,V> n) { node = n; }

    boolean isEmpty() { return node == null; }
    K key() { return node.key; }
    V value() {
      if (isEmpty()) return null;
      else return node.value; 
    }
    
    /** Requires: node != null.  
      * @return left subtree */
    RefNode<K,V> left() { return node.left; }
    
    /** Requires: node != null.  
      * @return left subtree */
    RefNode<K,V> right() { return node.right; }
    
    /** Sets node to n. */
    void set(Node<K,V> n) { node = n; }
    
    /** Sets the entry in this node to (k,v), creating a new node if necessary.  Does NOT adjust size of any associated
      * TreeMap. */
    void set(K k, V v) {
      if (isEmpty()) node = new Node<K,V>(k,v);
      else {
        node.key = k;
        node.value = v;
      }
    }
    
    /** Requires: node != null. Sets the key field of node to k. */
    void setKey(K k) { node.key = k; }
    
    /** Requires: node != null. Sets the value field of node to v. */
    void setValue(V v) { node.value = v; }
    
    /** Requires: node != null. Sets the left field of node to rn. */
    void setLeft(RefNode<K,V> rn) { node.left = rn; }
    
    /** Requires: node != null. Sets the right field of node to rn. */
    void setRight(RefNode<K,V> rn) { node.right = rn; }
    
    /** If this is non-empty, deletes the entry in this. Does not affect size in enclosing TreeMap. */
    V remove() {
      if (isEmpty()) throw new MapException("Attempted to remove a non-existent key");
      V oldVal = node.value;
      if (left().isEmpty()) node = right().node;
      else if (right().isEmpty()) node = left().node;
      else {
        Node<K,V> minNode = right().removeMin();
        assert minNode != null;
        node.key = minNode.key;
        node.value = minNode.value;
      }
      return oldVal;
    }
    
    /** Deletes the Node in this tree with the minimum key. 
      * @return the deleted Node. */
    private Node<K,V> removeMin() {
      assert node != null;
      RefNode<K,V> current = this;
      V returnVal = node.value;
      do {
        RefNode<K,V> next = current.left();
        if (next.isEmpty()) break;  // current is ref to minimum node
        current = next;
      } while (true);
      Node<K,V> minNode = current.node;
      current.set(minNode.right.node);  // argument may be null
      return minNode;
    }
    
    private static class Node<K extends Comparable<K>, V> {
      /* Fields */
      K key;
      V value;
      RefNode<K,V> left = makeEmpty();
      RefNode<K,V> right = makeEmpty();
      
      /* Constructor */
      Node(K k, V v) {
        key = k;
        value = v;
      }
    }
  }
}
class MapException extends RuntimeException {
  MapException(String s) { super(s); }
}
