/* A mutable list of elements of type E that can be scanned in either direction. */
interface BiListI<E> {
  
  /** @return true if this enumeration is empty. */
  boolean isEmpty();
  
  /** Constructs a new empty BiList. */
  BiListI<E> newList();
  
  /** @return the length of this list. */
  int length();
  
  /** Inserts object o at the front of this. 
    * @return this list. */
  BiListI<E> insertFront(E o);
  
  /** Inserts the object o at the rear of this list.
    * @return this list. */
  BiListI<E> insertRear(E o);
  
  /** Clears this;
    * @return this list. */
  BiList<E> clear();
  
  /** Removes the element f at the front of the list.
    * @return the removed element f. */
  E remFront();      
  
  /** Removes the last element l of this list. 
    * @return the removed element l.
    * @throws a BiListException if no such element exists. 
    */
  E remRear();
  
  /** @return a new iterator for traversing this. */
  BiIteratorI<E> newIterator(); 
}

interface BiIteratorI<E> {
  /** Moves cursor to the first element f of the collection.  Behavior unspecified on empty collecton. 
    * @return the first element f. */
  E first();
  /** Places the cursor on the last element l of the collection.  Behavior is unspecified if the collection is empty. 
    * @return the last element l. */
  E last(); 
  /** Moves cursor to the next element n of the collection.  Behavior unspecified on empty collecton. 
    * @return the next element n. */ 
  E next();
  /** Moves cursor to the previous element p of the collection.  Behavior is unspecified if the collection is empty. 
    * @return the next element p. */
  E prev(); 
  /** @return true if cursor is pointing to the dummy node (the header) beyond the last element of the collection. */
  boolean atEnd();
  /** @return true if cursor is pointing to the first element of the collection. */
  boolean atStart();
  /** @return the value of the current element (pointed to by the cursor). 
    * @throws IteratorException if no such element exists.
    */
  E currentItem();
  /** Destructively inserts the object o immediately before the current item.  If the list is empty, or the cursor 
    * is atEnd(), the node is inserted immediately before the end of the list. */
  void insert(E o); 
  /** Removes the current item.
    * @throws an IteratorException if the cursor is atEnd(). */
  E remove(); 
}

interface ILambda<A,B> {
  B apply(A host); 
}

/** Exception class for signalling errors in processing lists (implementing ListI) */
class ListException extends RuntimeException {
  ListException(String s) { super(s); }
}

/** Exception class for signalling errors in using an iterator */
class IteratorException extends RuntimeException {
  IteratorException(String s) { super(s); }
}

/** A circularly linked class implementing the BiListI interface. */
class BiList<E> implements BiListI<E> {
  
  /** The header node for the list. */
  private Node<E> head = new Node<E>();
  /** The length of this list. */
  private int length = 0;
  
  /* Relying on the default constructor for BiList. */
  
  /* Implementations of BiListI methods.  See BiListI for contracts. */
  public String toString(){
    StringBuffer s = new StringBuffer("(");
    for (BiIteratorI i = newIterator(); ! i.atEnd(); i.next()) {
      if (i.atStart()) s.append(i.currentItem().toString());
      else s.append(" " + i.currentItem().toString());
    }  
    return s.append(")").toString();
  }
  public boolean isEmpty() { return head.succ == head; } 
  public BiList<E> newList() { return new BiList<E>(); }
  public int length() { return length; }
  
  public BiList<E> clear() {
    head.pred = head;
    head.succ = head;
    length = 0;
    return this;
  }
  
  public <R> BiList<R> map(ILambda<E, R> f) {
    BiIteratorI<E> it = newIterator();
    BiList<R> result = new BiList<R>();
    while (! it.atEnd()) {  // atEnd() is true after going one next() beyond the first element
      E current = it.currentItem();
      System.err.println("inserting " + f.apply(current) + " in " + result);
      result.insertRear(f.apply(current));
      it.next();
    }
    return result;
  }

  public BiList<E> insertFront(E o) { 
    Node<E> next = head.succ;
    Node<E> newNode = new Node<E>(o,head,next);  // allocate new Node
    // insert new Node
    head.succ = newNode;
    next.pred = newNode;
    length++;
    return this;
  }
  
  public BiList<E> insertRear(E o)  { 
    Node<E> prev = head.pred;
    Node<E> newNode = new Node<E>(o,prev,head);  // allocate new Node
    // insert new Node
    head.pred = newNode;
    prev.succ = newNode;
    length++;
    return this;
  }
  
  public E remFront() {
    if (isEmpty()) 
      throw new BiListException("Attempt to remove first element of empty CircList");
    Node<E> next = head.succ;
    E remob = next.item;
    Node<E> newNext = next.succ;
    head.succ = newNext;
    newNext.pred = head;
    length--;
    
    return remob;
  }
  
  /* Virtual constructors for iterators; identical except for return type */
  public BiIteratorI<E> newIterator() { return new BiIterator(); }
  
  public E remRear() {
    if (isEmpty())
      throw new BiListException("Attempt to remove last element of empty CircList");
    Node<E> last = head.pred;
    E remob = last.item;
    Node<E> lastPred = last.pred;
    head.pred = lastPred;
    lastPred.succ = head;
    length--;
    return remob;
  }
  
  boolean delete(E elt) {
    for (BiIteratorI<E> i = newIterator(); !i.atEnd(); i.next())
      if (i.currentItem().equals(elt)) {
      i.remove();
      return true;
    };
    return false;
  }
  
  /* Static nested Node class.  Embedded as nested class to hide access. */
  private static class Node<E> {
    E item;
    Node<E> pred,succ;
    
    Node(E i, Node<E> p, Node<E> s) {
      item = i;
      pred = p;
      succ = s;
    }
    
    Node() {  // allocate header
      item = null;
      pred = this;
      succ = this;
    }
  } 
  
  /** A true inner class that includes a link (BiList.this) to the enclosing BiList. */
  private class BiIterator implements BiIteratorI<E> {
    
    Node<E> current;
    
    BiIterator() {
      current = head.succ;  // head is an abbreviation for BiList.this.head
    }
    
    public E first() { 
      current = head.succ;  // head is an abbreviation for BiList.this.head
      return current.item;
    }
    
    public E last() { 
      current = head.pred;  // head is an abbreviation for BiList.this.head
      return current.item;
    }
    
    public E next() {
      current = current.succ;        // wraps around end
      return current.item;
    }
    
    public E prev() {
      current = current.pred;        // wraps around end
      return current.item;
    }
    
    public E currentItem() { 
      if (current == head)           // head is an abbreviation for BiList.this.head
        throw new IteratorException("No current element in " + BiList.this);
      return current.item; 
    }
    
    public boolean atEnd() { return current == head; }  // head is an abbreviation for BiList.this.head
   
    /** Returns true if CircList is non-empty and current is the first node in the list. */
    public boolean atStart() { return ! atEnd() && current == BiList.this.head.succ; }
     
    public void insert(E o) {
      
      // pre: true
      // post: Node containing o is inserted before current item; current is unchanged
      
      Node<E> oldPred = current.pred;
      Node<E> newNode = new Node<E>(o, oldPred, current);  // allocate new node
      current.pred = newNode;                              // insert it
      oldPred.succ = newNode;
      BiList.this.length++;
      
    } 
    
    public E remove() {
      
      // pre:  current is valid
      // post: current becomes current.succ
      
      if (current == BiList.this.head) throw 
        new IteratorException("BiIterator.remove() applied at end of BiList " + BiList.this);
      Node<E> cPred = current.pred;
      Node<E> cSucc = current.succ;
      cPred.succ = cSucc;
      cSucc.pred = cPred;
      E remOb = current.item;
      current = cSucc;  
      BiList.this.length--;
      return remOb;
    } 
  }
  
  static class BiListException extends RuntimeException {
    BiListException(String s) { super(s); }
  }
}

class TestBiList {
  public static void main(String args[]) {
    BiList<Integer> c = new BiList<Integer>();
    System.out.println(c.isEmpty());
    System.out.println("c is: " + c);
    c.insertFront(new Integer(567));
    System.out.println("c is: " + c);
    System.out.println(c.isEmpty());
    c.insertRear(new Integer(72));
    System.out.println("c is: " + c);
    c.insertRear(new Integer(4330));
    System.out.println("c is: " + c);
    BiList<BiList<Integer>> d = new BiList<BiList<Integer>>();
    d.insertFront(c);
    System.out.println("d is: " + d);
  }
}